
#ifndef MP_IO_MANAGER_H_
#define MP_IO_MANAGER_H_

#include "stdio_synchronizer.hpp"
#include "self_pipe.hpp"
#include "tcp_socket.hpp"
#include "thread_pool.hpp"

#include <atomic>
#include <condition_variable>
#include <functional>
#include <cstdint>
#include <list>
#include <memory>
#include <mutex>
#include <string>
#include <vector>
#include <queue>
#include <thread>
#include <unordered_map>

#ifdef _WIN32
#include <Winsock2.h>
#else
#include <sys/select.h>
#include <unistd.h>
#endif /* _WIN32 */

#include <cstdint>
#include <stdexcept>
#include <iostream>
#include <string>
#include <memory>
#include <mutex>

// Note:
//   If it is using auto type, it will be easier to read because #ifdef lines are reduced.
//   #ifdef was caused by the difference of socket file descriptor type on Windows and Ubuntu.

namespace multiplatform {

// service that operates IO Handling.
// It polls sockets for input and output, processes read and write operations and calls the appropriate callbacks.
class io_manager {
public:

#ifdef _WIN32
  using socket_fd_t = SOCKET;
#else // _WIN32
  using socket_fd_t = int;
#endif // _WIN32

  io_manager(void): m_callback_workers(1){
    if(m_debug){ _debug_print("create io_manager"); }
    // Start worker after everything has been initialized
    m_poll_worker = std::thread(std::bind(&io_manager::poll, this));
  }
  ~io_manager(void){
    if(m_debug){ _debug_print("destroy io_manager"); }
    m_should_stop = true;
    m_notifier.notify();
    if(m_poll_worker.joinable()){ m_poll_worker.join(); }
    m_callback_workers.stop();
  }
  io_manager(const io_manager&) = delete;              // Removed copy constructor
  io_manager& operator=(const io_manager&) = delete;   // Removed assignment operator

  // reset number of io_manager workers assigned to this io_manager
  // this can be safely called at runtime, even if the io_manager is currently running
  // it can be useful if you need to re-adjust the number of workers
  void set_max_workers(std::size_t counts){
    m_callback_workers.set_max_threads(counts);
  }

  // track socket
  // add socket to io_manager tracking for read/write operation
  // socket is polled only if read or write callback is defined
  //
  // \param socket socket to be tracked
  // \param rd_callback callback to be executed on read event
  // \param wr_callback callback to be executed on write event
  void track(const tcp_socket& socket, const std::function<void(socket_fd_t)>& rd_callback = nullptr, const std::function<void(socket_fd_t)>& wr_callback = nullptr){
    if(m_debug){ _debug_print("track new socket"); }
    std::lock_guard<std::mutex> socket_lock(m_tracked_sockets_mutex);
    auto& track_info                    = m_tracked_sockets[socket.get_fd()];
    track_info.rd_callback              = rd_callback;
    track_info.wr_callback              = wr_callback;
    track_info.marked_for_untrack       = false;
    track_info.is_executing_rd_callback = false;
    track_info.is_executing_wr_callback = false;
    m_notifier.notify();
  }


  // Set/Update the read callback. If the socket has not yet been tracked, it will be tracked immediately.
  void set_rd_callback(const tcp_socket& socket, const std::function<void(socket_fd_t)>& read_callback){
    if(m_debug){ _debug_print("update read socket tracking callback"); }
    std::lock_guard<std::mutex> socket_lock(m_tracked_sockets_mutex);
    auto& track_info       = m_tracked_sockets[socket.get_fd()];
    track_info.rd_callback = read_callback;
    m_notifier.notify();
  }

  // Set/Update the write callback. If the socket has not yet been tracked, it will be tracked immediately.
  void set_wr_callback(const tcp_socket& socket, const std::function<void(socket_fd_t)>& write_callback){
    if(m_debug){ _debug_print("update write socket tracking callback"); }
    std::lock_guard<std::mutex> socket_lock(m_tracked_sockets_mutex);
    auto& track_info       = m_tracked_sockets[socket.get_fd()];
    track_info.wr_callback = write_callback;
    m_notifier.notify();
  }

  // remove socket from io_manager tracking
  // socket is marked for untracking and will effectively be removed asynchronously from tracking once
  //  * poll wakes up
  //  * no callback are being executed for that socket
  //
  // re-adding track while socket is pending for untrack is fine and will simply cancel the untrack operation
  //
  // \param socket socket to be untracked
  void untrack(const tcp_socket& socket){
    std::lock_guard<std::mutex> socket_lock(m_tracked_sockets_mutex);
    auto it = m_tracked_sockets.find(socket.get_fd());
    if(it == m_tracked_sockets.end()){ return; }
    if(it->second.is_executing_rd_callback || it->second.is_executing_wr_callback){
      if(m_debug){ _debug_print("mark socket for untracking"); }
      it->second.marked_for_untrack = true;
    }
    else{
      if(m_debug){ _debug_print("untrack socket"); }
      m_tracked_sockets.erase(it);
      m_wait_for_removal_condition.notify_all();
    }
    m_notifier.notify();
  }

  // wait until the socket has been effectively removed
  // basically wait until all pending callbacks are executed
  //
  // \param socket socket to wait for
  void wait_for_removal(const tcp_socket& socket){
    if(m_debug){ _debug_print("waiting for socket removal"); }
    std::unique_lock<std::mutex> socket_lock(m_tracked_sockets_mutex);
    m_wait_for_removal_condition.wait(socket_lock, [&](){
      if(m_debug){ _debug_print("socket has been removed"); }
      return m_tracked_sockets.find(socket.get_fd()) == m_tracked_sockets.end();
    });
  }

private:
  // struct tracked_socket
  // contains information about what a current socket is tracking
  //  * rd_callback: callback to be executed on read availability
  //  * is_executing_rd_callback: whether the rd callback is currently being executed or not
  //  * wr_callback: callback to be executed on write availability
  //  * is_executing_wr_callback: whether the wr callback is currently being executed or not
  //  * marked_for_untrack: whether the socket is marked for being untrack (that is, will be untracked whenever all the callback completed their execution)
  struct tracked_socket {
    tracked_socket(void): rd_callback(nullptr), wr_callback(nullptr){} // Create without callback registration.

    // read/write event callbacks
    std::function<void(socket_fd_t)> rd_callback;
    std::function<void(socket_fd_t)> wr_callback;
    std::atomic<bool> is_executing_rd_callback = ATOMIC_VAR_INIT(false);
    std::atomic<bool> is_executing_wr_callback = ATOMIC_VAR_INIT(false);

    // marked for untrack
    std::atomic<bool> marked_for_untrack = ATOMIC_VAR_INIT(false);
  };

  // poll worker function
  // main loop of the background thread in charge of the io_manager in charge of polling fds
  void poll(void){
    if(m_debug){ _debug_print("starting poll() worker"); }
    while(!m_should_stop){
      int ndfs = init_poll_fds_info();
      // setup timeout
      struct timeval* timeout_ptr = NULL;
#ifdef _MP_TIMEOUT
      struct timeval timeout;
      timeout.tv_usec = _MP_TIMEOUT;
      timeout_ptr     = &timeout;
#endif // _MP_TIMEOUT
      if(m_debug){ _debug_print("polling fds"); }
      if(select(ndfs, &m_rd_set, &m_wr_set, NULL, timeout_ptr) > 0){
        process_events();
      }
      else{
        if(m_debug){ _debug_print("poll woke up, but nothing to process"); }
      }
    }
    if(m_debug){ _debug_print("stop poll() worker"); }
  }

  // init m_poll_fds_info
  // simply initialize m_polled_fds variable based on m_tracked_sockets information
  //
  // \return maximum fd value polled
  int init_poll_fds_info(void){
    std::lock_guard<std::mutex> socket_lock(m_tracked_sockets_mutex);
    m_polled_fds.clear();
    FD_ZERO(&m_rd_set);
    FD_ZERO(&m_wr_set);
    int ndfs = (int) m_notifier.get_read_fd();
    FD_SET(m_notifier.get_read_fd(), &m_rd_set);
    m_polled_fds.push_back(m_notifier.get_read_fd());
    for(const auto& socket : m_tracked_sockets){
      const auto& fd          = socket.first;
      const auto& socket_info = socket.second;
      bool should_rd = socket_info.rd_callback && !socket_info.is_executing_rd_callback;
      if(should_rd){ FD_SET(fd, &m_rd_set); }
      bool should_wr = socket_info.wr_callback && !socket_info.is_executing_wr_callback;
      if(should_wr){ FD_SET(fd, &m_wr_set); }
      if(should_rd || should_wr || socket_info.marked_for_untrack){
        m_polled_fds.push_back(fd);
      }
      if((should_rd || should_wr) && (int) fd > ndfs){
        ndfs = (int) fd;
      }
    }
    return ndfs + 1;
  }

  // process poll detected events
  // called whenever select/poll completed to check read and write availablity
  void process_events(void){
    std::lock_guard<std::mutex> socket_lock(m_tracked_sockets_mutex);
    if(m_debug){ _debug_print("processing events"); }
    for(const auto& fd : m_polled_fds){
      if(fd == m_notifier.get_read_fd() && FD_ISSET(fd, &m_rd_set)){
        m_notifier.clear();
        continue;
      }
      auto it = m_tracked_sockets.find(fd);
      if(it == m_tracked_sockets.end()){ continue; }
      auto& socket = it->second;
      if(FD_ISSET(fd, &m_rd_set) && socket.rd_callback && !socket.is_executing_rd_callback){
        process_rd_event(fd, socket);
      }
      if(FD_ISSET(fd, &m_wr_set) && socket.wr_callback && !socket.is_executing_wr_callback){
        process_wr_event(fd, socket);
      }
      if(socket.marked_for_untrack && !socket.is_executing_rd_callback && !socket.is_executing_wr_callback){
        if(m_debug){ _debug_print("untrack socket"); }
        m_tracked_sockets.erase(it);
        m_wait_for_removal_condition.notify_all();
      }
    }
  }

  // process read event reported by select/poll for a given socket
  //
  // \param fd fd for which a read event has been reported
  // \param socket tracked_socket associated to the given fd
#ifdef _WIN32
  void process_rd_event(const SOCKET& fd, tracked_socket& socket)
#else // _WIN32
  void process_rd_event(const int& fd, tracked_socket& socket)
#endif // _WIN32
  {
    if(m_debug){ _debug_print("processing read event"); }
    auto rd_callback = socket.rd_callback;
    socket.is_executing_rd_callback = true;
    m_callback_workers << [=] {
      if(m_debug){ _debug_print("execute read callback"); }
      rd_callback(fd);
      std::lock_guard<std::mutex> socket_lock(m_tracked_sockets_mutex);
      auto it = m_tracked_sockets.find(fd);
      if(it == m_tracked_sockets.end()){ return; }
      auto& socket                    = it->second;
      socket.is_executing_rd_callback = false;
      if(socket.marked_for_untrack && !socket.is_executing_wr_callback){
        if(m_debug){ _debug_print("untrack socket"); }
        m_tracked_sockets.erase(it);
        m_wait_for_removal_condition.notify_all();
      }
      m_notifier.notify();
    };
  }

  // process write event reported by select/poll for a given socket
  //
  // \param fd fd for which a write event has been reported
  // \param socket tracked_socket associated to the given fd
#ifdef _WIN32
  void process_wr_event(const SOCKET& fd, tracked_socket& socket)
#else // _WIN32
  void process_wr_event(const int& fd, tracked_socket& socket)
#endif // _WIN32
  {
    if(m_debug){ _debug_print("processing write event"); }
    auto wr_callback = socket.wr_callback;
    socket.is_executing_wr_callback = true;
    m_callback_workers << [=] {
      if(m_debug){ _debug_print("execute write callback"); }
      wr_callback(fd);
      std::lock_guard<std::mutex> socket_lock(m_tracked_sockets_mutex);
      auto it = m_tracked_sockets.find(fd);
      if(it == m_tracked_sockets.end()){ return; }
      auto& socket                    = it->second;
      socket.is_executing_wr_callback = false;
      if(socket.marked_for_untrack && !socket.is_executing_rd_callback){
        if(m_debug){ _debug_print("untrack socket"); }
        m_tracked_sockets.erase(it);
        m_wait_for_removal_condition.notify_all();
      }
      m_notifier.notify();
    };
  }

  // tracked sockets
#ifdef _WIN32
  std::unordered_map<SOCKET, tracked_socket> m_tracked_sockets;
#else // _WIN32
  std::unordered_map<int, tracked_socket> m_tracked_sockets;
#endif // _WIN32

  // whether the worker should stop or not
  std::atomic<bool> m_should_stop = ATOMIC_VAR_INIT(false);

  // poll thread
  std::thread m_poll_worker;

  // callback workers
  thread_pool m_callback_workers;

  // thread safety
  std::mutex m_tracked_sockets_mutex;

  // data structure given to select (list of fds to poll)
#ifdef _WIN32
  std::vector<SOCKET> m_polled_fds;
#else // _WIN32
  std::vector<int> m_polled_fds;
#endif // _WIN32

  fd_set m_rd_set;
  fd_set m_wr_set;

  // condition variable to wait on removal
  std::condition_variable m_wait_for_removal_condition;

  // fd associated to the pipe used to wake up the poll call
  multiplatform::self_pipe m_notifier;

  std::atomic<bool> m_debug = ATOMIC_VAR_INIT(false);
};

static std::shared_ptr<io_manager> io_manager_default_instance = nullptr;

inline const std::shared_ptr<io_manager>& get_default_io_manager(void){
  if(io_manager_default_instance == nullptr){
    io_manager_default_instance = std::make_shared<io_manager>();
  }
  return io_manager_default_instance;
}

inline void set_default_io_manager(const std::shared_ptr<io_manager>& instance){
  //DEBUG//_debug_print("setting new default_io_manager");
  io_manager_default_instance = instance;
}

} // namespace multiplatform

#endif // MP_IO_MANAGER_H_
