
#ifndef MP_SELF_PIPE_H_
#define MP_SELF_PIPE_H_

#ifdef _WIN32
#include <Winsock2.h>
#include <stdio.h>
#pragma comment(lib, "Ws2_32.lib")
#else // _WIN32
#include <fcntl.h>  // for unistd.h
#include <unistd.h> // for pipe(), read(), write()
#endif // _WIN32

#include <cstdint>
#include <stdexcept>
#include <iostream>
#include <string>
#include <memory>
#include <mutex>

namespace multiplatform {

// This class will act as a wrapper to describe "self-pipe trick" coding the as same as linux way on Windows.
class self_pipe {
public:
#ifdef _WIN32
  self_pipe(void): m_loopback_fd(INVALID_SOCKET){
    // Create a server
    m_loopback_fd = ::socket(AF_INET, SOCK_DGRAM, 0);
    if(m_loopback_fd == INVALID_SOCKET){ _error_throw("fail socket()"); }
    u_long flags = 1;
    ioctlsocket(m_loopback_fd, FIONBIO, &flags);
    // Bind server to localhost
    {
      struct sockaddr_in inaddr;
      memset(&inaddr, 0, sizeof(inaddr));
      inaddr.sin_family      = AF_INET;
      inaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
      inaddr.sin_port        = 0;
      if(bind(m_loopback_fd, (struct sockaddr*) &inaddr, sizeof(inaddr)) == SOCKET_ERROR){ _error_throw("fail bind()"); }
    }
    // Retrieve server information
    m_addr_len = sizeof(m_addr);
    memset(&m_addr, 0, sizeof(m_addr));
    if(getsockname(m_loopback_fd, &m_addr, &m_addr_len) == SOCKET_ERROR){ _error_throw("fail getsockname()"); }
    // connect read fd to the server
    if(connect(m_loopback_fd, &m_addr, m_addr_len) == SOCKET_ERROR){ _error_throw("fail connect()"); }
  }
  ~self_pipe(void){
    if(m_loopback_fd != INVALID_SOCKET){ closesocket(m_loopback_fd); }
  }
#else // _WIN32
  self_pipe(void): m_pipe_fds{-1, -1}{
    if(pipe(m_pipe_fds) == -1){ _error_throw("pipe() failure"); }
  }
  ~self_pipe(void){
    if(m_pipe_fds[0] != -1){ close(m_pipe_fds[0]); }
    if(m_pipe_fds[1] != -1){ close(m_pipe_fds[1]); }
  }
#endif // _WIN32

  self_pipe(const self_pipe&) = delete;             // Removed copy constructor.
  self_pipe& operator=(const self_pipe&) = delete;  // Removed assignment operator.

#ifdef _WIN32
  SOCKET get_read_fd(void) const{ return m_loopback_fd; }
  SOCKET get_write_fd(void) const{ return m_loopback_fd; }
#else // _WIN32
  int get_read_fd(void) const{ return m_pipe_fds[0]; }
  int get_write_fd(void) const{ return m_pipe_fds[1]; }
#endif // _WIN32

  // This method converts from asynchronous events to file events and notifying them.
  void notify(void){
#ifdef _WIN32
    (void) sendto(m_loopback_fd, "d", 1, 0, &m_addr, m_addr_len);
#else // _WIN32
    (void) write(m_pipe_fds[1], "d", 1);
#endif // _WIN32
  }
  // This method clears the received event.
  void clear(void){
    char buf[1024];
#ifdef _WIN32
    (void) recvfrom(m_loopback_fd, buf, 1024, 0, &m_addr, &m_addr_len);
#else // _WIN32
    (void) read(m_pipe_fds[0], buf, 1024);
#endif // _WIN32
  }

private:
  //bool m_debug;            // Print debug messages for this class.
#ifdef _WIN32
  SOCKET m_loopback_fd;    // Loopback socket file descriptor for pipe. (Share with read/write)
  struct sockaddr m_addr;  // Socket information.
  int m_addr_len;          // Socket information. This is sizeof(m_addr).
#else // _WIN32
  int m_pipe_fds[2];       // File descriptors for pipe. ([0]:read side, [1]:write side)
#endif // _WIN32
};

} // namespace multiplatform

#endif // MP_SELF_PIPE_H_

