
#ifndef MP_STDIO_SYNCHRONIZER_H_
#define MP_STDIO_SYNCHRONIZER_H_

// Use a long macro name to avoid duplication with existing macros.
#define _MACRO__STRINGIFY(n) #n
#define _MACRO__TO_STRING(n) _MACRO__STRINGIFY(n)

#define _GROBAL_DEBUG false

#include <atomic>
//#include <condition_variable>
//#include <functional>
//#include <cstdint>
//#include <list>
//#include <memory>
//#include <mutex>
#include <iostream>
#include <sstream>
#include <string>
//#include <vector>
//#include <queue>
//#include <thread>
//#include <unordered_map>

//#include <cstdint>
//#include <stdexcept>
//#include <iostream>
//#include <string>
//#include <memory>
#include <mutex>

// Note:
//   If it is using auto type, it will be easier to read because #ifdef lines are reduced.
//   #ifdef was caused by the difference of socket file descriptor type on Windows and Ubuntu.

namespace multiplatform {

// multiplatform::stdio_synchronizer is the class providing asynchronously data get/set interfaces
// and synchronously stdio control.
//
// behavior specifications of the standard output:
// - This class can request printing instructions asynchronously in a multithreaded application,
//   and prevent output contents from being mixed.
// - The print request is never locked, and if output is impossible, it is temporarily queued in the buffer.
// - In this class it is guaranteed to keep the time series when the print request was made.
class stdio_synchronizer {
public:
  // Constant values::
  const int error_level_expr = 0x1;
  const int warn_level_expr  = 0x2;
  const int info_level_expr  = 0x4;
  const int debug_level_expr = 0x8;

  const int error_flag_expr = 0x1;
  const int warn_flag_expr  = 0x3;
  const int info_flag_expr  = 0x7;
  const int debug_flag_expr = 0xF;

  stdio_synchronizer(void):
      m_default_istream(&std::cin),
      m_default_ostream(&std::cerr){
    m_export_flag = info_flag_expr;
  }
  ~stdio_synchronizer(void){
  }
  stdio_synchronizer(const stdio_synchronizer&) = delete;            // Removed copy constructor
  stdio_synchronizer& operator=(const stdio_synchronizer&) = delete; // Removed assignment operator

  // Comparison operator
  //bool operator==(const stdio_synchronizer& rhs) const{ return m_type == rhs.m_type; }
  //bool operator!=(const stdio_synchronizer& rhs) const{ return m_type != rhs.m_type; }

  inline void change_level_error(void){ m_export_flag = error_flag_expr; }
  inline void change_level_warn(void){  m_export_flag = warn_flag_expr; }
  inline void change_level_info(void){  m_export_flag = info_flag_expr; }
  inline void change_level_debug(void){ m_export_flag = debug_flag_expr; }

  inline void set_level_error_only(void){ m_export_flag |= error_level_expr; }
  inline void set_level_warn_only(void){  m_export_flag |= warn_level_expr; }
  inline void set_level_info_only(void){  m_export_flag |= info_level_expr; }
  inline void set_level_debug_only(void){ m_export_flag |= debug_level_expr; }

  inline void unset_level_error_only(void){ m_export_flag &= ~error_level_expr; }
  inline void unset_level_warn_only(void){  m_export_flag &= ~warn_level_expr; }
  inline void unset_level_info_only(void){  m_export_flag &= ~info_level_expr; }
  inline void unset_level_debug_only(void){ m_export_flag &= ~debug_level_expr; }

  inline void reset_level(void){ m_export_flag = 0; }

  inline int get_level(void){ return m_export_flag; }

  inline void raw_print(const std::string& msg){
    std::lock_guard<std::mutex> msg_lock(m_write_mutex);
    (*m_default_ostream) << msg;
  }
  inline void raw_print_with_endl(const std::string& msg){
    std::lock_guard<std::mutex> msg_lock(m_write_mutex);
    (*m_default_ostream) << msg << std::endl;
  }

  inline void error_throw(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
    if(get_level() & error_level_expr){
      if(enable_debug){
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[ERROR][" << filename << ":" << linenum << "] " << msg << std::endl;
        throw std::runtime_error(msg);
      }
      else{
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[ERROR]" << " " << msg << std::endl;
        throw std::runtime_error(msg);
      }
    }
  }
  inline void error_throw(const std::string& msg, const std::string& filename, const std::string& linenum){
    error_throw(msg, false, filename, linenum);
  }
  inline void error_print(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
    if(get_level() & error_level_expr){
      if(enable_debug){
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[ERROR][" << filename << ":" << linenum << "] " << msg << std::endl;
      }
      else{
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[ERROR]" << " " << msg << std::endl;
      }
    }
  }
  inline void error_print(const std::string& msg, const std::string& filename, const std::string& linenum){
    error_print(msg, false, filename, linenum);
  }
  inline void warning_print(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
    if(get_level() & warn_level_expr){
      if(enable_debug){
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[WARNING][" << filename << ":" << linenum << "] " << msg << std::endl;
      }
      else{
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[WARNING]" << " " << msg << std::endl;
      }
    }
  }
  inline void warning_print(const std::string& msg, const std::string& filename, const std::string& linenum){
    warning_print(msg, false, filename, linenum);
  }
  inline void info_print(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
    if(get_level() & info_level_expr){
      if(enable_debug){
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[INFO][" << filename << ":" << linenum << "] " << msg << std::endl;
      }
      else{
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[INFO]" << " " << msg << std::endl;
      }
    }
  }
  inline void info_print(const std::string& msg, const std::string& filename, const std::string& linenum){
    info_print(msg, false, filename, linenum);
  }
  inline void debug_print(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
    if(get_level() & debug_level_expr){
      if(enable_debug){
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[DEBUG][" << filename << ":" << linenum << "] " << msg << std::endl;
      }
      else{
        std::lock_guard<std::mutex> msg_lock(m_write_mutex);
        (*m_default_ostream) << "[DEBUG]" << " " << msg << std::endl;
      }
    }
  }
  inline void debug_print(const std::string& msg, const std::string& filename, const std::string& linenum){
    debug_print(msg, false, filename, linenum);
  }

  inline void msg_scan(std::string& msg){
    std::lock_guard<std::mutex> msg_lock(m_read_mutex);
    std::cin >> msg;
  }

private:
  std::atomic<int> m_export_flag;    // print flags
  std::istream *m_default_istream;   // std::cin default
  std::ostream *m_default_ostream;   // std::cerr default
  std::mutex m_read_mutex;           // std::cin mutex
  std::mutex m_write_mutex;          // std::cout, std::cerr and std::clog mutex
};

//extern std::unique_ptr<multiplatform::stdio_synchronizer> default_synchronizer;
//default_synchronizer = nullptr;

static std::unique_ptr<multiplatform::stdio_synchronizer> default_synchronizer = nullptr;

inline void raw_print(const std::string& msg){
  if(default_synchronizer){
    default_synchronizer->raw_print(msg);
  }
}
inline void raw_print_with_endl(const std::string& msg){
  if(default_synchronizer){
    default_synchronizer->raw_print_with_endl(msg);
  }
}

inline void error_throw(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->error_throw(msg, enable_debug, filename, linenum);
  }
}
inline void error_throw(const std::string& msg, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->error_throw(msg, filename, linenum);
  }
}
inline void error_print(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->error_print(msg, enable_debug, filename, linenum);
  }
}
inline void error_print(const std::string& msg, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->error_print(msg, filename, linenum);
  }
}
inline void warning_print(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->warning_print(msg, enable_debug, filename, linenum);
  }
}
inline void warning_print(const std::string& msg, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->warning_print(msg, filename, linenum);
  }
}
inline void debug_print(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->debug_print(msg, enable_debug, filename, linenum);
  }
}
inline void debug_print(const std::string& msg, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->debug_print(msg, filename, linenum);
  }
}
inline void info_print(const std::string& msg, bool enable_debug, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->info_print(msg, enable_debug, filename, linenum);
  }
}
inline void info_print(const std::string& msg, const std::string& filename, const std::string& linenum){
  if(default_synchronizer){
    default_synchronizer->info_print(msg, filename, linenum);
  }
}

inline void msg_scan(std::string& msg){
  if(default_synchronizer){
    default_synchronizer->msg_scan(msg);
  }
}

#define _init_default_synchronizer multiplatform::default_synchronizer.reset(new multiplatform::stdio_synchronizer());
#define _raw_print(msg) multiplatform::raw_print(msg);
#define _raw_print_with_endl(msg) multiplatform::raw_print_with_endl(msg);
#define _error_throw(msg) multiplatform::error_throw(msg, _GROBAL_DEBUG, __FILE__, _MACRO__TO_STRING(__LINE__));
#define _error_print(msg) multiplatform::error_print(msg, _GROBAL_DEBUG, __FILE__, _MACRO__TO_STRING(__LINE__));
#define _warning_print(msg) multiplatform::warning_print(msg, _GROBAL_DEBUG, __FILE__, _MACRO__TO_STRING(__LINE__));
#define _debug_print(msg) multiplatform::debug_print(msg, _GROBAL_DEBUG, __FILE__, _MACRO__TO_STRING(__LINE__));
#define _info_print(msg) multiplatform::info_print(msg, _GROBAL_DEBUG, __FILE__, _MACRO__TO_STRING(__LINE__));

} // namespace multiplatform

#endif // MP_STDIO_SYNCHRONIZER_H_

