
#ifndef MP_TCP_SOCKET_H_
#define MP_TCP_SOCKET_H_

#ifdef _WIN32
#pragma comment( lib, "ws2_32.lib")
#endif // _WIN32

#include "stdio_synchronizer.hpp"
#include "uuid_generator.hpp"
#include <atomic>
#include <condition_variable>
#include <functional>
#include <cstdint>
#include <list>
#include <memory>
#include <mutex>
#include <string>
#include <vector>
#include <queue>
#include <thread>
#include <unordered_map>

#ifdef _MSC_VER
#include <basetsd.h>
#else // _MSC_VER
#endif // _MSC_VER

#ifdef _WIN32
#include <Winsock2.h>
#include <WS2tcpip.h>
#else
#include <sys/select.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <unistd.h>
#endif /* _WIN32 */

//#include <cstdint>
#include <stdexcept>
#include <iostream>
#include <cstring>
//#include <memory>
//#include <mutex>

// Note:
//   If it is using auto type, it will be easier to read because #ifdef lines are reduced.
//   #ifdef was caused by the difference of socket file descriptor type on Windows and Ubuntu.

namespace multiplatform {

#ifdef _WIN32
  constexpr SOCKET INVALID_SOCKET_EXPR = INVALID_SOCKET;
  constexpr bool IS_UNIX_EXPR = false;
#else // _WIN32
  constexpr int INVALID_SOCKET_EXPR = -1;
  constexpr bool IS_UNIX_EXPR = true;
#endif // _WIN32

// multiplatform::tcp_socket is the class providing low-level TCP socket features.
// The tcp_socket provides a simple but convenient abstraction to unix and windows sockets.
// It also provides a socket type checker to ensure that server-only operations are only processable on server sockets, and client-only operations are only processable on client sockets.
class tcp_socket {
public:

#ifdef _MSC_VER
  using ssize_t = SSIZE_T;
#endif

#ifdef _WIN32
  using socket_fd_t = SOCKET;
#else // _WIN32
  using socket_fd_t = int;
#endif // _WIN32

  // possible types of a TCP socket, either a client or a server
  // type is used to prevent the used of client specific operations on a server socket (and vice-versa)
  //
  // UNKNOWN is used when socket type could not be determined for now
  enum class type {
    CLIENT,
    SERVER,
    UNKNOWN
  };

  tcp_socket(void): m_fd(INVALID_SOCKET_EXPR), m_host("") , m_port(0) , m_type(type::UNKNOWN){
    multiplatform::uuid uuid;
    m_uuid = uuid.v4();
    if(m_debug){ _debug_print("create tcp_socket"); }
  }
  ~tcp_socket(void) = default;

  // custom ctor
  // build socket from existing file descriptor
  //
  // \param fd fd of the raw socket that will be used to init the tcp_socket object
  // \param host host associated to the socket
  // \param port port associated to the socket
  // \param t type of the socket (client or server)
  tcp_socket(socket_fd_t fd, const std::string& host, std::uint32_t port, type t):
      m_fd(fd), m_host(host), m_port(port), m_type(t){
    multiplatform::uuid uuid;
    m_uuid = uuid.v4();
    if(m_debug){ _debug_print("create tcp_socket"); }
  }

  // Move constructor
  tcp_socket(tcp_socket&& socket):
      m_uuid(std::move(socket.m_uuid)), m_fd(std::move(socket.m_fd)), m_host(socket.m_host), m_port(socket.m_port), m_type(socket.m_type){
    multiplatform::uuid uuid;
    socket.m_uuid = uuid.v4();
    socket.m_fd   = INVALID_SOCKET_EXPR;
    socket.m_type = type::UNKNOWN;
    if(m_debug){ _debug_print("moved tcp_socket"); }
  }
  tcp_socket(const tcp_socket&) = delete;             // Removed copy constructor
  tcp_socket& operator=(const tcp_socket&) = delete;  // Removed assignment operator

  // Comparison operator
  bool operator==(const tcp_socket& rhs) const{ return m_fd == rhs.m_fd && m_type == rhs.m_type; }
  bool operator!=(const tcp_socket& rhs) const{ return m_fd != rhs.m_fd || m_type != rhs.m_type; }

  // Read data synchronously from the underlying socket.
  // The socket must be of type client to process this operation. If the type of the socket is unknown, the socket type will be set to client.
  //
  // \param size_to_read Number of bytes to read (might read less than requested)
  // \return Returns the read bytes
  std::vector<char> recv(std::size_t size_to_read){
    prepare_socket();
    try_set_type(type::CLIENT);
    std::vector<char> data(size_to_read, 0);
#ifdef _WIN32
    ssize_t rd_size = ::recv(m_fd, const_cast<char*>(data.data()), static_cast<int>(size_to_read), 0);
    if(rd_size == SOCKET_ERROR){ _error_throw("recv() failure"); }
#else // _WIN32
    ssize_t rd_size = ::recv(m_fd, const_cast<char*>(data.data()), size_to_read, 0);
    if(rd_size == -1){ _error_throw("recv() failure"); }
#endif // _WIN32
    if(rd_size == 0){ _error_throw("nothing to read, socket has been closed by remote host"); }
    data.resize(rd_size);
    return data;
  }

  // Send data synchronously to the underlying socket.
  // The socket must be of type client to process this operation. If the type of the socket is unknown, the socket type will be set to client.
  //
  // \param data Buffer containing bytes to be written
  // \param size_to_write Number of bytes to send
  // \return Returns the number of bytes that were effectively sent.
  std::size_t send(const std::vector<char>& data, std::size_t size_to_write){
    prepare_socket();
    try_set_type(type::CLIENT);
#ifdef _WIN32
    ssize_t wr_size = ::send(m_fd, data.data(), static_cast<int>(size_to_write), 0);
    if(wr_size == SOCKET_ERROR){ _error_throw("send() failure"); }
#else // _WIN32
    ssize_t wr_size = ::send(m_fd, data.data(), size_to_write, 0);
    if(wr_size == -1){ _error_throw("send() failure"); }
#endif // _WIN32
    return wr_size;
  }

  // Server socket operation: connect()
  // Connect the socket to the remote server.
  // The socket must be of type client to process this operation. If the type of the socket is unknown, the socket type will be set to client.
  //
  // \param host Hostname of the target server
  // \param port Port of the target server
  // \param timeout_msecs maximum time to connect (will block until connect succeed or timeout expire). 0 will block undefinitely. If timeout expires, connection fails
  void connect(const std::string& host, std::uint32_t port, std::uint32_t timeout_msecs = 0){
    // Reset host and port
    m_host = host;
    m_port = port;

    prepare_socket();
    try_set_type(type::CLIENT);

    struct sockaddr_storage ss;
    socklen_t addr_len;

    std::memset(&ss, 0, sizeof(ss));

#ifdef _WIN32
    // This case that incorrect #ifdef processing was done or there was a problem with usage.
    if(IS_UNIX_EXPR && (m_port == 0)){
      _error_throw("unix domain socket is not supported on windows");
    }
#else // _WIN32
    // Handle case of unix sockets if port is 0.
    if(IS_UNIX_EXPR && (m_port == 0)){
      struct sockaddr_un* addr = reinterpret_cast<struct sockaddr_un*>(&ss);
      strncpy(addr->sun_path, host.c_str(), sizeof(addr->sun_path) - 1);
      ss.ss_family = AF_UNIX;
      addr_len     = sizeof(*addr);
    }
#endif // _WIN32
    else if(is_ipv6()){
      struct sockaddr_in6* addr = reinterpret_cast<struct sockaddr_in6*>(&ss);
      if(::inet_pton(AF_INET6, host.data(), &addr->sin6_addr) < 0){ _error_throw("inet_pton() failure"); }
      ss.ss_family    = AF_INET6;
      addr->sin6_port = htons(port);
      addr_len        = sizeof(*addr);
    }
    else{ // With the advent of IPv6, it is necessary to use getaddrinfo and freeaddrinfo for DNS resolution.
      struct addrinfo* response = nullptr;
      struct addrinfo hints;
      memset(&hints, 0, sizeof(hints));
      hints.ai_socktype = SOCK_STREAM;
      hints.ai_family   = AF_INET;
      if(getaddrinfo(host.c_str(), nullptr, &hints, &response) != 0){ _error_throw("getaddrinfo() failure"); }
      struct sockaddr_in* addr = reinterpret_cast<struct sockaddr_in*>(&ss);
      addr->sin_addr = ((struct sockaddr_in*) (response->ai_addr))->sin_addr;
      addr->sin_port = htons(port);
      ss.ss_family   = AF_INET;
      addr_len       = sizeof(*addr);
      freeaddrinfo(response);
    }

    if(timeout_msecs > 0){
      // for timeout connection handling:
      //  1. set socket to non blocking
      //  2. connect
      //  3. poll select
      //  4. check connection status
#ifdef _WIN32
      u_long mode = 1;
      if(ioctlsocket(m_fd, FIONBIO, &mode) != 0)
#else // _WIN32
      if(fcntl(m_fd, F_SETFL, fcntl(m_fd, F_GETFL, 0) | O_NONBLOCK) == -1)
#endif // _WIN32
      {
        close();
        _error_throw("connect() set non-blocking failure");
      }
    }
    else{
      // If timeout is not set, it need to check that the socket is in blocking mode.
#ifdef _WIN32
      u_long mode = 0;
      if(ioctlsocket(m_fd, FIONBIO, &mode) != 0)
#else // _WIN32
      if(fcntl(m_fd, F_SETFL, fcntl(m_fd, F_GETFL, 0) & (~O_NONBLOCK)) == -1)
#endif // _WIN32
      {
        close();
        _error_throw("connect() set blocking failure");
      }
    }

    int ret = ::connect(m_fd, reinterpret_cast<const struct sockaddr*>(&ss), addr_len);
#ifdef _WIN32
    if(ret == -1 && WSAGetLastError() != WSAEWOULDBLOCK)
#else // _WIN32
    if(ret < 0 && errno != EINPROGRESS)
#endif // _WIN32
    {
      close();
      _error_throw("connect() failure");
    }

    if(timeout_msecs > 0){
      timeval tv;
      tv.tv_sec  = (timeout_msecs / 1000);
      tv.tv_usec = ((timeout_msecs - (tv.tv_sec * 1000)) * 1000);

      fd_set set;
      FD_ZERO(&set);
      FD_SET(m_fd, &set);

      // 1 means we are connected.
      // 0/-1 means a timeout.
#ifdef _WIN32
      if(select(static_cast<int>(m_fd) + 1, NULL, &set, NULL, &tv) == 1)
#else // _WIN32
      if(select(m_fd + 1, NULL, &set, NULL, &tv) == 1)
#endif // _WIN32
      {
        // Make sure there are no async connection errors
        int err       = 0;
#ifdef _WIN32
        int len       = sizeof(int);
        if(getsockopt(m_fd, SOL_SOCKET, SO_ERROR, reinterpret_cast<char*>(&err), &len) == -1 || err != 0)
#else // _WIN32
        socklen_t len = sizeof(socket_fd_t);
        if(getsockopt(m_fd, SOL_SOCKET, SO_ERROR, &err, &len) == -1 || err != 0)
#endif // _WIN32
        {
          close();
          _error_throw("connect() failure"); 
        }

        // Set back to blocking mode as the user of this class is expecting
#ifdef _WIN32
        u_long mode = 0;
        if(ioctlsocket(m_fd, FIONBIO, &mode) != 0)
#else // _WIN32
        if(fcntl(m_fd, F_SETFL, fcntl(m_fd, F_GETFL, 0) & (~O_NONBLOCK)) == -1)
#endif // _WIN32
        {
          close();
          _error_throw("connect() set blocking failure"); 
        }
      }
      else{
        close();
        _error_throw("connect() timed out"); 
      }
    }
  }


  // Server socket operation: bind()
  // Binds the socket to the given host and port.
  // The socket must be of type server to process this operation. If the type of the socket is unknown, the socket type will be set to server.
  //
  // \param host Hostname to be bind to
  // \param port Port to be bind to
  void bind(const std::string& host, std::uint32_t port){
    // Reset host and port
    m_host = host;
    m_port = port;

    prepare_socket();
    try_set_type(type::SERVER);

    struct sockaddr_storage ss;
    socklen_t addr_len;

    std::memset(&ss, 0, sizeof(ss));  // 0-init addr info struct

#ifdef _WIN32
    // This case that incorrect #ifdef processing was done or there was a problem with usage.
    if(IS_UNIX_EXPR && (m_port == 0)){
      _error_throw("unix domain socket is not supported on windows");
    }
#else // _WIN32
    // Handle case of unix sockets if port is 0.
    if(IS_UNIX_EXPR && (m_port == 0)){
      struct sockaddr_un* addr = reinterpret_cast<struct sockaddr_un*>(&ss);
      strncpy(addr->sun_path, host.c_str(), sizeof(addr->sun_path) - 1);
      ss.ss_family = AF_UNIX;
      addr_len     = sizeof(*addr);
    }
#endif // _WIN32
    else if(is_ipv6()){
      struct sockaddr_in6* addr = reinterpret_cast<struct sockaddr_in6*>(&ss);
      if(::inet_pton(AF_INET6, host.data(), &addr->sin6_addr) < 0){ _error_throw("inet_pton() failure");  }
      addr->sin6_port = htons(port);
      ss.ss_family    = AF_INET6;
      addr_len        = sizeof(*addr);
    }
    else{ // With the advent of IPv6, it is necessary to use getaddrinfo and freeaddrinfo for DNS resolution.
      struct addrinfo* response = nullptr;
      if(getaddrinfo(host.c_str(), nullptr, nullptr, &response) != 0){ _error_throw("getaddrinfo() failure"); }
      struct sockaddr_in* addr = reinterpret_cast<struct sockaddr_in*>(&ss);
      addr->sin_addr = ((struct sockaddr_in*) (response->ai_addr))->sin_addr;
      addr->sin_port = htons(port);
      ss.ss_family   = AF_INET;
      addr_len       = sizeof(*addr);
      freeaddrinfo(response);
    }
#ifdef _WIN32
    if(::bind(m_fd, reinterpret_cast<const struct sockaddr*>(&ss), addr_len) == SOCKET_ERROR){ _error_throw("bind() failure"); }
#else // _WIN32
    if(::bind(m_fd, reinterpret_cast<const struct sockaddr*>(&ss), addr_len) == -1){ _error_throw("bind() failure"); }
#endif // _WIN32
  }

  // Server socket operation: listen()
  // Make the socket listen for incoming connections.
  // Socket must be of type server to process this operation. If the type of the socket is unknown, the socket type will be set to server.
  //
  // \param max_connection_queue Size of the queue for incoming connections to be processed by the server
  void listen(std::size_t max_connection_queue){
    prepare_socket();
    try_set_type(type::SERVER);
#ifdef _WIN32
    if(::listen(m_fd, static_cast<int>(max_connection_queue)) == SOCKET_ERROR){ _error_throw("listen() failure"); }
#else // _WIN32
    if(::listen(m_fd, max_connection_queue) == -1){ _error_throw("listen() failure"); }
#endif // _WIN32
  }

  // Server socket operation: accept()
  // Accept a new incoming connection.
  // The socket must be of type server to process this operation. If the type of the socket is unknown, the socket type will be set to server.
  //
  // \return Return the tcp_socket associated to the newly accepted connection.
  tcp_socket accept(void){
    prepare_socket();
    try_set_type(type::SERVER);
    struct sockaddr_storage ss;
    socklen_t addrlen = sizeof(ss);
    socket_fd_t client_fd = ::accept(m_fd, reinterpret_cast<struct sockaddr*>(&ss), &addrlen);
    if(client_fd == INVALID_SOCKET_EXPR){ _error_throw("accept() failure"); }
    //! now determine host and port based on socket type
    std::string saddr;
    std::uint32_t port;
    if(ss.ss_family == AF_INET6){
      struct sockaddr_in6* addr6 = reinterpret_cast<struct sockaddr_in6*>(&ss);
      char buf[INET6_ADDRSTRLEN] = {};
      const char* addr           = ::inet_ntop(ss.ss_family, &addr6->sin6_addr, buf, INET6_ADDRSTRLEN);
      if(addr){
        saddr = std::string("[") + addr + "]";
      }
      port = ntohs(addr6->sin6_port);
    }
    else{
      struct sockaddr_in* addr4 = reinterpret_cast<struct sockaddr_in*>(&ss);
      char buf[INET_ADDRSTRLEN] = {};
      const char* addr          = ::inet_ntop(ss.ss_family, &addr4->sin_addr, buf, INET_ADDRSTRLEN);
      if(addr){
        saddr = addr;
      }
      port = ntohs(addr4->sin_port);
    }
    return {client_fd, saddr, port, type::CLIENT};
  }

  // Close the underlying socket.
  void close(void){
    if(m_fd != INVALID_SOCKET_EXPR){
    if(m_debug){ _debug_print("close socket"); }
#ifdef _WIN32
      ::closesocket(m_fd);
#else // _WIN32
      ::close(m_fd);
#endif // _WIN32
    }
    m_fd = INVALID_SOCKET_EXPR;
    m_type = type::UNKNOWN;
  }

  // Getter
  const std::string& get_uuid(void) const{ return m_uuid; }
  const std::string& get_host(void) const{ return m_host; }
  std::uint32_t get_port(void) const{ return m_port; }
  type get_type(void) const{ return m_type; }

  // set type, should be used if some operations determining socket type
  // have been done on the behalf of the tcp_socket instance
  //
  // \param t type of the socket
  void set_type(type t){ m_type = t; }

  // direct access to the underlying fd
  //
  // \return underlying socket fd
  socket_fd_t get_fd(void) const{ return m_fd; }

  // \return whether the host is IPV6
  bool is_ipv6(void) const{ return m_host.find(':') != std::string::npos; }

private:
  // create a new socket if no socket has been initialized yet
  void prepare_socket(void){
    if(m_fd != INVALID_SOCKET_EXPR){ return; }

    if(IS_UNIX_EXPR && (m_port == 0)){
#ifdef _WIN32
      _error_throw("unix domain socket is not supported on windows");
#else // _WIN32
      m_fd = socket(AF_UNIX, SOCK_STREAM, 0);
#endif // _WIN32
    }
    else if(is_ipv6()){
      m_fd = socket(AF_INET6, SOCK_STREAM, 0);
    }
    else{
      m_fd = socket(AF_INET, SOCK_STREAM, 0);
    }
    m_type = type::UNKNOWN;

    if(m_fd == INVALID_SOCKET_EXPR){ _error_throw("tcp_socket::prepare_socket: socket() failure"); }
  }

  void try_set_type(type t){
    // type can be changed only when type is type::UNKNOWN.
    if(m_type != type::UNKNOWN && m_type != t){
      _error_throw("trying to perform invalid operation on socket"); 
    }
    m_type = t;
  }

  std::string m_uuid;   // unique identifier of the socket
  socket_fd_t m_fd;     // fd associated to the socket
  std::string m_host;   // socket hostname information
  std::uint32_t m_port; // socket port information
  type m_type;          // type of the socket
  std::mutex m_mutex;   // mutex for std::cout and std::cerr
  std::atomic<bool> m_debug = ATOMIC_VAR_INIT(false);
};

} // namespace multiplatform

#endif // MP_TCP_SOCKET_H_

