// recognizer.h

#ifdef _MSC_VER
#define NOMINMAX
#endif

#include "recognizer1.h"

namespace kaldi {
namespace nnet1 {
NnetRecognizer::NnetRecognizer(
        const kaldi::nnet1::Nnet& nnet,
        kaldi::nnet1::PdfPrior& pdf_prior,
        const kaldi::TransitionModel& trans_model,
        kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt,
        std::shared_ptr<kaldi::OnlineFeatureMfcnet>& features,
        std::shared_ptr<kaldi::OnlineFeatureInterface>& features_if,
        const fst::Fst<fst::StdArc>* decode_fst,
        const kaldi::LatticeFasterDecoderConfig& decoder_config,
        const fst::SymbolTable* phone_syms,
        const fst::SymbolTable* word_syms,
        const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
        std::string& lm_name,
        int emulate_julius,
        int nbest_count,
        int progress_output_interval,
        std::shared_ptr<result_server> result_client):
    id_(features->GetID()),
    trans_model_(trans_model),
    features_(features), features_if_(features_if),
    decodable_(nnet, pdf_prior, trans_model, opt, features_if),
    decoder_(*decode_fst, decoder_config),
    phone_syms_(phone_syms), word_syms_(word_syms),
    lexicon_info_(lexicon_info),
    result_client_(result_client),
    lm_name_(lm_name),
    emulate_julius_(emulate_julius),
    nbest_count_(nbest_count),
    progress_output_interval_(progress_output_interval),
    features_input_is_finished(false){
  //if(features_){ std::string id_ = features_->GetID(); }
  // remove path prefix
#ifdef _MSC_VER
  std::string::size_type pos = lm_name.rfind("\\");
#else
  std::string::size_type pos = lm_name.rfind("/");
#endif
  lm_name_ = (pos != std::string::npos) ? lm_name.substr(++pos) : lm_name;
  emulate_julius_ = (emulate_julius != 0) ? 1 : 0;
  delimiter_sym_ = (emulate_julius_) ? ".\n" : "";
  //DEBUG//std::cout << "fst ::: " << (decode_fst ? "not NULL" : "NULL") << std::endl;
}

std::shared_ptr<OnlineFeatureInterface> NnetRecognizer::GetOnlineFeatureInterface()
{
  return features_if_;
}

void NnetRecognizer::Recognize()
{
  
  //DEBUG//std::cout << "frames ::: " <<  features_if_->NumFramesReady() << std::endl;
  //DEBUG//std::cout << "decoder ::: " <<  decoder_.NumFramesDecoded() << std::endl;
  //DEBUG//std::cout << "decodable ::: " << decodable_.NumFramesReady() << std::endl;
  if(features_){
    features_->Initialize();
    std::ostringstream ss;
    ss << std::fixed << std::setprecision(6)
       << "<SOURCEINFO SOURCEID=\"" << features_->GetSourceInfo().source_id
       <<           "\" AZIMUTH=\"" << features_->GetSourceInfo().azimuth
       <<         "\" ELEVATION=\"" << features_->GetSourceInfo().elevation
       <<               "\" SEC=\"" << features_->GetSourceInfo().tv_sec
       <<              "\" USEC=\"" << features_->GetSourceInfo().tv_usec << "\" />\n" << delimiter_sym_
       << "<STARTRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    result_client_->broadcast(ss.str());
  }

  //DEBUG//KALDI_LOG << "started recognize key id=" << id_;
  Timer timer; timer.Reset();
  //decoder_.Decode(&decodable_);
  decoder_.InitDecoding();

  int progress_iframes = 0;
  int progress_dframes = progress_output_interval_;
  while(true){
    if(features_){
      //DEBUG//KALDI_LOG << "receiving features key id=" << id_;
      features_->Receive();
    }
    if(features_if_->IsLastFrame(features_if_->NumFramesReady() - 1)){
      //while(true){
      //  if(decodable_.NumFramesReady() == features_if_->NumFramesReady()){ break; }
      //  multiplatform::time::ms_sleep(10);
      //}
      decoder_.AdvanceDecoding(&decodable_);
      //DEBUG//KALDI_LOG << "final frame recognized key id=" << id_;
      break;
    }
    if(progress_output_interval_ > 0){
      if(features_if_->NumFramesReady() > progress_iframes){
        decoder_.AdvanceDecoding(&decodable_, progress_output_interval_);
        //DEBUG//KALDI_LOG << "progress frame recognized key id=" << id_;
        progress_iframes += progress_output_interval_;
      }
      if(decoder_.NumFramesDecoded() >= progress_dframes){
        WriteIntermediateResult(decoder_.NumFramesDecoded());
        progress_dframes += progress_output_interval_;
      }
      multiplatform::time::ms_sleep(5);
    }
    else{
      multiplatform::time::ms_sleep(500);
    }
  }

  decoder_.FinalizeDecoding();
  //DEBUG//KALDI_LOG << "finalized decoding key id=" << id_;
  double elapsed = timer.Elapsed();
  KALDI_LOG << elapsed << " sec, RT=" << (elapsed * 100.0 / features_if_->NumFramesReady());

  if(features_){
    std::ostringstream ss;
    ss << "<ENDRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    result_client_->broadcast(ss.str());
    WriteResult();
  }
  PrintResult(std::cout);
  if(features_){
    result_client_->add_recogend_key(id_);
  }
}

void NnetRecognizer::WriteIntermediateResult(kaldi::int32 frame)
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    std::ostringstream ss;
    ss << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    result_client_->broadcast(ss.str());
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat, false);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::ostringstream ss;
    ss << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    if(lm_name_ != ""){ ss << " NAME=\"" << lm_name_ << "\""; }
    ss << std::fixed << std::setprecision(6)
       << ">\n"
       << " <PHYPO RANK=\"" << 1
       <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
       <<   "\" AMSCORE=\"" << -(double)weight.Value2()
       <<   "\" LMSCORE=\"" << -(double)weight.Value1()
       <<     "\" FRAME=\"" << frame
       <<      "\" TIME=\"" << time(NULL)
       << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word = string_tokenizer(classid, "+")[0];
      //std::string word = (word_index < string_tokenizer(classid, "+").size()) ? string_tokenizer(classid, "+")[word_index] : string_tokenizer(classid, "+")[0];
      ////std::string word;
      ////std::vector<std::string> tokens = string_tokenizer(classid, "+");
      ////switch(tokens.size()){
      ////case 2:  word = tokens[1]; break;
      ////case 1:
      ////case 4:
      ////default: word = tokens[0]; break;
      ////}
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << std::fixed << std::setprecision(3)
         << "  <WHYPO WORD=\"" << word
         <<    "\" CLASSID=\"" << classid
         <<      "\" PHONE=\"" << ""
         <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    ss << " </PHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    result_client_->broadcast(ss.str());
  }
}

void NnetRecognizer::WriteResult()
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    std::ostringstream ss;
    ss << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    result_client_->broadcast(ss.str());
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::ostringstream ss;
    ss << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    if(lm_name_ != ""){ ss << " NAME=\"" << lm_name_ << "\""; }
    ss << std::fixed << std::setprecision(6)
       << ">\n"
       << " <SHYPO RANK=\"" << 1
       <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
       <<   "\" AMSCORE=\"" << -(double)weight.Value2()
       <<   "\" LMSCORE=\"" << -(double)weight.Value1() << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word = string_tokenizer(classid, "+")[0];
      //std::string word = (word_index < string_tokenizer(classid, "+").size()) ? string_tokenizer(classid, "+")[word_index] : string_tokenizer(classid, "+")[0];
      ////std::string word;
      ////std::vector<std::string> tokens = string_tokenizer(classid, "+");
      ////switch(tokens.size()){
      ////case 2:  word = tokens[1]; break;
      ////case 1:
      ////case 4:
      ////default: word = tokens[0]; break;
      ////}
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << std::fixed << std::setprecision(3)
         << "  <WHYPO WORD=\"" << word
         <<    "\" CLASSID=\"" << classid
         <<      "\" PHONE=\"" << ""
         <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    ss << " </SHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    result_client_->broadcast(ss.str());
  }
}

void NnetRecognizer::PrintResult(std::ostream& out)
{
  std::stringstream ss;
  if(decoder_.NumFramesDecoded() <= 0){
    if(emulate_julius_){
      ss << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n"
         << "### Recognition: 2nd pass (RL heuristic best-first) \n"
         << "STAT: 00 \n"
         << "sentence1: \n"
         << "wseq1: \n"
         << "phseq1: \n"
         << "cmscore1: \n"
         << "score1: \n\n"
         << "ERROR: recognition failed.";
      out << ss.str() << std::endl;
    }
    else{
      ss << "recognition failed\n"
         << "result: \n"
         << "phone_seq: ";
      out << ss.str() << std::endl;
    }
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned, &weight);

  if(emulate_julius_){
    if(features_){
      ss << std::fixed << std::setprecision(6)
         <<   "source_id = " << features_->GetSourceInfo().source_id
         <<   ", azimuth = " << features_->GetSourceInfo().azimuth
         << ", elevation = " << features_->GetSourceInfo().elevation
         <<       ", sec = " << features_->GetSourceInfo().tv_sec
         <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
    }
    ss << "### Recognition: 2nd pass (RL heuristic best-first) \n"
       << "STAT: 00 \n"
       << "sentence1: ";
    if(word_syms_ != NULL){
      size_t i = 0;
      while(i < words.size()){
        std::string classid = word_syms_->Find(words[i]);
        ss << string_tokenizer(classid, "+")[0] << " ";
        if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
        i++;
      }
    }
    else{
      size_t i = 0;
      while(i < words.size()){ ss << words[i++] << " "; }
    }
    ss << "\n";
  }

  ss << (emulate_julius_ ? "wseq1: " : "result: ");
  if(word_syms_ != NULL){
    size_t i = 0;
    while(i < words.size()){
      std::string s = word_syms_->Find(words[i]);
      if(s == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
  }
  else{
    size_t i = 0;
    while(i < words.size()){ ss << words[i++] << " "; }
  }
  ss << "\n";

  ss << (emulate_julius_ ? "phseq1: " : "phone_seq: ");
  if(phone_syms_ != NULL){
    size_t i = 0;
    while(i < phones_aligned.size()){
      kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
      kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
      std::string s = phone_syms_->Find(phone);
      if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
  }
  ss << "\n";

  if(emulate_julius_){
    ss << "cmscore1: ";
    size_t i = 0;
    while(i < words.size()){
      ss << std::fixed << std::setprecision(3) << 1.0f << " ";
      i++;
    }
    ss << "\n";
    ss << std::fixed << std::setprecision(6)
       << "score1: " << -(double)(weight.Value1() + weight.Value2())
       << " ( AM: " << -(double)weight.Value2() << ", LM: " << -(double)weight.Value1() << " )";
  }
  else{
    ss << std::fixed << std::setprecision(6)
       <<     "score=" << -(double)(weight.Value1() + weight.Value2())
       << ", amscore=" << -(double)weight.Value2()
       << ", lmscore=" << -(double)weight.Value1();
  }
  out << ss.str() << std::endl;

  /*
  Lattice lat;
  decoder.GetRawLattice(&lat);
  CompactLattice clat;
  if(!DeterminizeLatticePhonePrunedWrapper(trans_model, &lat, decoder_config.lattice_beam, &clat, decoder_config.det_opts)){ }
  CompactLattice best_path_clat;
  CompactLatticeShortestPath(clat, &best_path_clat);

  Lattice best_path_lat;
  ConvertLattice(best_path_clat, &best_path_lat);
  */
}
} // namespace nnet1
} // namespace kaldi

