// recognizer.h

#ifdef _MSC_VER
#define NOMINMAX
#endif

#include "recognizer3.h"

namespace kaldi {
namespace nnet3 {
NnetRecognizer::NnetRecognizer(
        kaldi::nnet3::AmNnetSimple& nnet,
        kaldi::nnet3::AmNnetSimple& cmb_nnet,
        const kaldi::TransitionModel& trans_model,
        const kaldi::TransitionModel& cmb_trans_model,
        kaldi::nnet3::NnetSimpleLoopedComputationOptions opt,
        kaldi::nnet3::DecodableNnetSimpleLoopedInfo &decodable_info,
        kaldi::nnet3::DecodableNnetSimpleLoopedInfo &cmb_decodable_info,
        kaldi::OnlineNnet2FeaturePipelineInfo *feature_info,
        std::shared_ptr<kaldi::OnlineFeatureMfcnet>& features,
        std::shared_ptr<kaldi::OnlineFeatureInterface>& features_if,
        const fst::Fst<fst::StdArc>* decode_fst,
        const fst::Fst<fst::StdArc>* cmb_decode_fst,
        const kaldi::LatticeFasterDecoderConfig& decoder_config,
        const fst::SymbolTable* phone_syms,
        const fst::SymbolTable* cmb_phone_syms,
        const fst::SymbolTable* word_syms,
        const fst::SymbolTable* cmb_word_syms,
        const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
        const kaldi::WordAlignLatticeLexiconInfo& cmb_lexicon_info,
        std::string& lm_name,
        int emulate_julius,
        int nbest_count,
        int progress_output_interval,
        std::vector<float>& combine_lat_weights,
        bool enable_print_result,
        bool enable_print_progress,
        bool enable_print_sentence,
        bool enable_print_lat_combine,
        bool enable_lat_combine,
        std::shared_ptr<result_server> result_client):
    id_(features->GetID()),
    trans_model_(trans_model), cmb_trans_model_(cmb_trans_model), opt_(opt),
    decodable_info_(decodable_info), cmb_decodable_info_(cmb_decodable_info),
    feature_info_(feature_info), features_(features), features_if_(features_if),
    decoder_(*decode_fst, decoder_config), cmb_decoder_(*cmb_decode_fst, decoder_config),
    phone_syms_(phone_syms), cmb_phone_syms_(cmb_phone_syms),
    word_syms_(word_syms), cmb_word_syms_(cmb_word_syms),
    lexicon_info_(lexicon_info), cmb_lexicon_info_(cmb_lexicon_info),
    result_client_(result_client),
    lm_name_(lm_name),
    emulate_julius_(emulate_julius),
    nbest_count_(nbest_count),
    progress_output_interval_(progress_output_interval),
    combine_lat_weights_(combine_lat_weights),
    enable_print_result_(enable_print_result),
    enable_print_progress_(enable_print_progress),
    enable_print_sentence_(enable_print_sentence),
    enable_print_lat_combine_(enable_print_lat_combine),
    enable_lat_combine_(enable_lat_combine),
    features_input_is_finished(false){
  //if(features_){ std::string id_ = features_->GetID(); }
  // remove path prefix
#ifdef _MSC_VER
  std::string::size_type pos = lm_name.rfind("\\");
#else
  std::string::size_type pos = lm_name.rfind("/");
#endif
  lm_name_ = (pos != std::string::npos) ? lm_name.substr(++pos) : lm_name;
  emulate_julius_ = (emulate_julius != 0) ? 1 : 0;
  delimiter_sym_ = (emulate_julius_) ? ".\n" : "";

  adaptation_state = new kaldi::OnlineIvectorExtractorAdaptationState(feature_info->ivector_extractor_info);
  silence_weighting = new kaldi::OnlineSilenceWeighting(trans_model, feature_info->silence_weighting_config, opt.frame_subsampling_factor);
  feature_pipeline = new kaldi::OnlineNnet2FeaturePipelineMod(*feature_info, features_if_);
  feature_pipeline->SetAdaptationState(*adaptation_state);

  decodable_ = new kaldi::nnet3::DecodableNnet3SimpleOnlineForHark(decodable_info_, trans_model_, opt_, feature_pipeline->InputFeature(), feature_pipeline->IvectorFeature());
  cmb_decodable_ = new kaldi::nnet3::DecodableNnet3SimpleOnlineForHark(cmb_decodable_info_, cmb_trans_model_, opt_, feature_pipeline->InputFeature(), feature_pipeline->IvectorFeature());
}

std::shared_ptr<OnlineFeatureInterface> NnetRecognizer::GetOnlineFeatureInterface()
{
  return features_if_;
}

void NnetRecognizer::Recognize()
{
  if(features_){
    features_->Initialize();
    std::ostringstream ss;
    ss << std::fixed << std::setprecision(6)
       << "<SOURCEINFO SOURCEID=\"" << features_->GetSourceInfo().source_id
       <<           "\" AZIMUTH=\"" << features_->GetSourceInfo().azimuth
       <<         "\" ELEVATION=\"" << features_->GetSourceInfo().elevation
       <<               "\" SEC=\"" << features_->GetSourceInfo().tv_sec
       <<              "\" USEC=\"" << features_->GetSourceInfo().tv_usec << "\" />\n" << delimiter_sym_
       << "<STARTRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    result_client_->broadcast(ss.str());
  }

  //DEBUG//KALDI_LOG << "started recognize key id=" << id_;
  Timer timer; timer.Reset();
  //decoder_.Decode(decodable_);
  decoder_.InitDecoding();

  int progress_iframes = 0;
  int progress_dframes = progress_output_interval_;
  while(true){
    if(features_){
      //DEBUG//KALDI_LOG << "receiving features key id=" << id_;
      features_->Receive();
    }
    if(silence_weighting->Active() && feature_pipeline->IvectorFeature() != NULL){
      silence_weighting->ComputeCurrentTraceback(decoder_);
      silence_weighting->GetDeltaWeights(feature_pipeline->NumFramesReady(), &delta_weights);
      feature_pipeline->IvectorFeature()->UpdateFrameWeights(delta_weights);
    }
    if(features_if_->IsLastFrame(features_if_->NumFramesReady() - 1)){
      //while(true){
      //  if(decodable_.NumFramesReady() == features_if_->NumFramesReady()){ break; }
      //  multiplatform::time::ms_sleep(10);
      //}
      decoder_.AdvanceDecoding(decodable_);
      //DEBUG//KALDI_LOG << "final frame recognized key id=" << id_;
      break;
    }
    if(progress_output_interval_ > 0){
      if(features_if_->NumFramesReady() > progress_iframes){
        decoder_.AdvanceDecoding(decodable_, progress_output_interval_);
        //DEBUG//KALDI_LOG << "progress frame recognized key id=" << id_;
        progress_iframes += progress_output_interval_;
      }
      if(decoder_.NumFramesDecoded() >= progress_dframes){
        WriteIntermediateResult(decoder_.NumFramesDecoded());
        progress_dframes += progress_output_interval_;
      }
      multiplatform::time::ms_sleep(5);
    }
    else{
      multiplatform::time::ms_sleep(500);
    }
  }

  decoder_.FinalizeDecoding();
  //DEBUG//KALDI_LOG << "finalized decoding key id=" << id_;
  double elapsed = timer.Elapsed();
  KALDI_LOG << elapsed << " sec, RT=" << (elapsed * 100.0 / features_if_->NumFramesReady());

  if(features_){
    std::ostringstream ss;
    ss << "<ENDRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    result_client_->broadcast(ss.str());
    WriteResult();
  }
  if(enable_print_result_ && enable_print_sentence_){
    PrintResult(std::cout);
  }

  if(features_){
    if(enable_lat_combine_){
      timer.Reset();
      cmb_decoder_.InitDecoding();
      cmb_decoder_.AdvanceDecoding(cmb_decodable_);
      cmb_decoder_.FinalizeDecoding();
      elapsed = timer.Elapsed();
      KALDI_LOG << elapsed << " sec, RT=" << (elapsed * 100.0 / features_if_->NumFramesReady());
      WriteRescoringResult();
      if(enable_print_result_ && enable_print_lat_combine_){
        PrintRescoringResult(std::cout);
      }
    }
  }

  if(features_){
    result_client_->add_recogend_key(id_);
  }
}

void NnetRecognizer::WriteIntermediateResult(kaldi::int32 frame)
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    std::ostringstream ss;
    ss << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    result_client_->broadcast(ss.str());
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat, false);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::ostringstream ss;
    ss << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    if(lm_name_ != ""){ ss << " NAME=\"" << lm_name_ << "\""; }
    ss << std::fixed << std::setprecision(6)
       << ">\n"
       << " <PHYPO RANK=\"" << 1
       <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
       <<   "\" AMSCORE=\"" << -(double)weight.Value2()
       <<   "\" LMSCORE=\"" << -(double)weight.Value1()
       <<     "\" FRAME=\"" << frame
       <<      "\" TIME=\"" << time(NULL)
       << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word = string_tokenizer(classid, "+")[0];
      //std::string word = (word_index < string_tokenizer(classid, "+").size()) ? string_tokenizer(classid, "+")[word_index] : string_tokenizer(classid, "+")[0];
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << std::fixed << std::setprecision(3)
         << "  <WHYPO WORD=\"" << word
         <<    "\" CLASSID=\"" << classid
         <<      "\" PHONE=\"" << ""
         <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    ss << " </PHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    result_client_->broadcast(ss.str());
  }
}

void NnetRecognizer::WriteResult()
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    std::ostringstream ss;
    ss << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    result_client_->broadcast(ss.str());
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::ostringstream ss;
    ss << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    if(lm_name_ != ""){ ss << " NAME=\"" << lm_name_ << "\""; }
    ss << std::fixed << std::setprecision(6)
       << ">\n"
       << " <SHYPO RANK=\"" << 1
       <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
       <<   "\" AMSCORE=\"" << -(double)weight.Value2()
       <<   "\" LMSCORE=\"" << -(double)weight.Value1() << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word = string_tokenizer(classid, "+")[0];
      //std::string word = (word_index < string_tokenizer(classid, "+").size()) ? string_tokenizer(classid, "+")[word_index] : string_tokenizer(classid, "+")[0];
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << std::fixed << std::setprecision(3)
         << "  <WHYPO WORD=\"" << word
         <<    "\" CLASSID=\"" << classid
         <<      "\" PHONE=\"" << ""
         <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    ss << " </SHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    result_client_->broadcast(ss.str());
  }
}

void NnetRecognizer::WriteRescoringResult()
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0 || cmb_decoder_.NumFramesDecoded() <= 0){
    std::ostringstream ss;
    ss << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    result_client_->broadcast(ss.str());
    return;
  }

  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  if(enable_lat_combine_){
    kaldi::Lattice lat1;
    kaldi::Lattice lat2;
    kaldi::CompactLattice clat1;
    kaldi::CompactLattice clat2;

    decoder_.GetRawLattice(&lat1);
    cmb_decoder_.GetRawLattice(&lat2);
    fst::ConvertLattice(lat1, &clat1);
    fst::ConvertLattice(lat2, &clat2);

    {
      float acoustic_scale = 1.0f; // temporary
      float lm_scale = 1.0f;       // temporary
      std::vector<std::vector<double> > lat_scale = fst::LatticeScale(lm_scale, acoustic_scale);
      fst::ScaleLattice(lat_scale, &clat1);
      fst::ScaleLattice(lat_scale, &clat2);
    }

    {
      bool fail = false;
      //float lat_weight = 0.5f; // lattice weights are half and half.
      //if(lat_weight > 0.0f && lat_weight < 1.0f){
      if(combine_lat_weights_.size() == 2){
        // Check lattices
        if(((&clat1)->Properties(fst::kTopSorted, false) == 0) && (fst::TopSort(&clat1) == false)){
          fail = true; // Cycles detected in clat1. Cannot normalize.
        }
        else if(((&clat2)->Properties(fst::kTopSorted, false) == 0) && (fst::TopSort(&clat2) == false)){
          fail = true; // Cycles detected in clat2. Cannot normalize.
        }
        else{}
        // Normalize lattice 1
        if(!fail){
          std::vector<double> beta;
          if(ComputeCompactLatticeBetas(clat1, &beta)){
            kaldi::CompactLattice::Arc::StateId start = (&clat1)->Start();
            kaldi::BaseFloat total_backward_cost = beta[start];
            total_backward_cost -= Log(combine_lat_weights_[0]);
            for(fst::StateIterator<CompactLattice> sit(clat1); !sit.Done(); sit.Next()){
              CompactLatticeWeight f = (&clat1)->Final(sit.Value());
              LatticeWeight w = f.Weight();
              w.SetValue1(w.Value1() + total_backward_cost);
              f.SetWeight(w);
              (&clat1)->SetFinal(sit.Value(), f);
            }
          }
          else{
            fail = true; // Failed to compute backward probabilities on clat1.
          }
        }
        // Normalize lattice 2
        if(!fail){
          std::vector<double> beta;
          if(ComputeCompactLatticeBetas(clat2, &beta)){
            kaldi::CompactLattice::Arc::StateId start = (&clat2)->Start();
            kaldi::BaseFloat total_backward_cost = beta[start];
            total_backward_cost -= Log(combine_lat_weights_[1]);
            for(fst::StateIterator<CompactLattice> sit(clat2); !sit.Done(); sit.Next()){
              CompactLatticeWeight f = (&clat2)->Final(sit.Value());
              LatticeWeight w = f.Weight();
              w.SetValue1(w.Value1() + total_backward_cost);
              f.SetWeight(w);
              (&clat2)->SetFinal(sit.Value(), f);
            }
          }
          else{
            fail = true; // Failed to compute backward probabilities on clat2.
          }
        }
      }
      else{
        fail = true; // lattice weight size must be 2.
        //// old comment // Lattice weights are positive numbers, and Lattice 1 weight + Lattice 2 weight must be 1.0f.
      }
      if(fail){
        std::ostringstream ss;
        ss << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
        result_client_->broadcast(ss.str());
        return;
      }
      fst::Union(&clat1, clat2);
    }
    //fst::ConvertLattice(clat1, &rescored_lat);
    kaldi::MinimumBayesRisk mbr(clat1);
    words = mbr.GetOneBest();
    weight.SetValue1(-0.0f);
    weight.SetValue2(-0.0f);
  }
  else if(false){ // This mode just outputs the offline results of the second model.
    kaldi::Lattice best_path_lat;
    cmb_decoder_.GetBestPath(&best_path_lat);
    GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
    GetWordAlignmentSequence(best_path_lat, cmb_trans_model_, cmb_lexicon_info_, &words, &weight);
  }
  else{
    return;
  }

  {
    std::ostringstream ss;
    ss << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    if(lm_name_ != ""){ ss << " NAME=\"" << lm_name_ << "\""; }
    ss << std::fixed << std::setprecision(6)
       << ">\n"
       << " <RHYPO RANK=\"" << 1
       <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
       <<   "\" AMSCORE=\"" << -(double)weight.Value2()
       <<   "\" LMSCORE=\"" << -(double)weight.Value1() << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word = string_tokenizer(classid, "+")[0];
      //std::string word = (word_index < string_tokenizer(classid, "+").size()) ? string_tokenizer(classid, "+")[word_index] : string_tokenizer(classid, "+")[0];
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << std::fixed << std::setprecision(3)
         << "  <WHYPO WORD=\"" << word
         <<    "\" CLASSID=\"" << classid
         <<      "\" PHONE=\"" << ""
         <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    ss << " </RHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    result_client_->broadcast(ss.str());
  }
}

void NnetRecognizer::PrintResult(std::ostream& out)
{
  std::stringstream ss;
  if(decoder_.NumFramesDecoded() <= 0){
    if(emulate_julius_){
      ss << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n"
         << "### Recognition: 2nd pass (RL heuristic best-first) \n"
         << "STAT: 00 \n"
         << "sentence1: \n"
         << "wseq1: \n"
         << "phseq1: \n"
         << "cmscore1: \n"
         << "score1: \n\n"
         << "ERROR: recognition failed.";
      out << ss.str() << std::endl;
    }
    else{
      ss << "recognition failed\n"
         << "result: \n"
         << "phone_seq: ";
      out << ss.str() << std::endl;
    }
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned, &weight);

  if(emulate_julius_){
    if(features_){
      ss << std::fixed << std::setprecision(6)
         <<   "source_id = " << features_->GetSourceInfo().source_id
         <<   ", azimuth = " << features_->GetSourceInfo().azimuth
         << ", elevation = " << features_->GetSourceInfo().elevation
         <<       ", sec = " << features_->GetSourceInfo().tv_sec
         <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
    }
    ss << "### Recognition: 2nd pass (RL heuristic best-first) \n"
       << "STAT: 00 \n"
       << "sentence1: ";
    if(word_syms_ != NULL){
      size_t i = 0;
      while(i < words.size()){
        std::string classid = word_syms_->Find(words[i]);
        ss << string_tokenizer(classid, "+")[0] << " ";
        if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
        i++;
      }
    }
    else{
      size_t i = 0;
      while(i < words.size()){ ss << words[i++] << " "; }
    }
    ss << "\n";
  }

  ss << (emulate_julius_ ? "wseq1: " : "result: ");
  if(word_syms_ != NULL){
    size_t i = 0;
    while(i < words.size()){
      std::string s = word_syms_->Find(words[i]);
      if(s == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
  }
  else{
    size_t i = 0;
    while(i < words.size()){ ss << words[i++] << " "; }
  }
  ss << "\n";

  ss << (emulate_julius_ ? "phseq1: " : "phone_seq: ");
  if(phone_syms_ != NULL){
    size_t i = 0;
    while(i < phones_aligned.size()){
      kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
      kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
      std::string s = phone_syms_->Find(phone);
      if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
  }
  ss << "\n";

  if(emulate_julius_){
    ss << "cmscore1: ";
    size_t i = 0;
    while(i < words.size()){
      ss << std::fixed << std::setprecision(3) << 1.0f << " ";
      i++;
    }
    ss << "\n";
    ss << std::fixed << std::setprecision(6)
       << "score1: " << -(double)(weight.Value1() + weight.Value2())
       << " ( AM: " << -(double)weight.Value2() << ", LM: " << -(double)weight.Value1() << " )";
  }
  else{
    ss << std::fixed << std::setprecision(6)
       <<     "score=" << -(double)(weight.Value1() + weight.Value2())
       << ", amscore=" << -(double)weight.Value2()
       << ", lmscore=" << -(double)weight.Value1();
  }
  out << ss.str() << std::endl;

  /*
  Lattice lat;
  decoder.GetRawLattice(&lat);
  CompactLattice clat;
  if(!DeterminizeLatticePhonePrunedWrapper(trans_model, &lat, decoder_config.lattice_beam, &clat, decoder_config.det_opts)){ }
  CompactLattice best_path_clat;
  CompactLatticeShortestPath(clat, &best_path_clat);

  Lattice best_path_lat;
  ConvertLattice(best_path_clat, &best_path_lat);
  */
}

void NnetRecognizer::PrintRescoringResult(std::ostream& out)
{
  std::stringstream ss;
  if(decoder_.NumFramesDecoded() <= 0){
    if(emulate_julius_){
      ss << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n"
         << "### Recognition: Rescoring pass \n"
         << "STAT: 00 \n"
         << "sentence1: \n"
         << "wseq1: \n"
         << "phseq1: \n"
         << "cmscore1: \n"
         << "score1: \n\n"
         << "ERROR: recognition failed.";
      out << ss.str() << std::endl;
    }
    else{
      ss << "recognition failed\n"
         << "result: \n"
         << "phone_seq: ";
      out << ss.str() << std::endl;
    }
    return;
  }

  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  if(enable_lat_combine_){
    kaldi::Lattice lat1;
    kaldi::Lattice lat2;
    kaldi::CompactLattice clat1;
    kaldi::CompactLattice clat2;

    decoder_.GetRawLattice(&lat1);
    cmb_decoder_.GetRawLattice(&lat2);
    fst::ConvertLattice(lat1, &clat1);
    fst::ConvertLattice(lat2, &clat2);

    {
      float acoustic_scale = 1.0f; // temporary
      float lm_scale = 1.0f;       // temporary
      std::vector<std::vector<double> > lat_scale = fst::LatticeScale(lm_scale, acoustic_scale);
      fst::ScaleLattice(lat_scale, &clat1);
      fst::ScaleLattice(lat_scale, &clat2);
    }

    {
      bool fail = false;
      float lat_weight = 0.5f; // lattice weights are half and half.
      if(lat_weight > 0.0f && lat_weight < 1.0f){
        // Check lattices
        if(((&clat1)->Properties(fst::kTopSorted, false) == 0) && (fst::TopSort(&clat1) == false)){
          fail = true; // Cycles detected in clat1. Cannot normalize.
        }
        else if(((&clat2)->Properties(fst::kTopSorted, false) == 0) && (fst::TopSort(&clat2) == false)){
          fail = true; // Cycles detected in clat2. Cannot normalize.
        }
        else{}
        // Normalize lattice 1
        if(!fail){
          std::vector<double> beta;
          if(ComputeCompactLatticeBetas(clat1, &beta)){
            kaldi::CompactLattice::Arc::StateId start = (&clat1)->Start();
            kaldi::BaseFloat total_backward_cost = beta[start];
            total_backward_cost -= Log(lat_weight);
            for(fst::StateIterator<CompactLattice> sit(clat1); !sit.Done(); sit.Next()){
              CompactLatticeWeight f = (&clat1)->Final(sit.Value());
              LatticeWeight w = f.Weight();
              w.SetValue1(w.Value1() + total_backward_cost);
              f.SetWeight(w);
              (&clat1)->SetFinal(sit.Value(), f);
            }
          }
          else{
            fail = true; // Failed to compute backward probabilities on clat1.
          }
        }
        // Normalize lattice 2
        if(!fail){
          std::vector<double> beta;
          if(ComputeCompactLatticeBetas(clat2, &beta)){
            kaldi::CompactLattice::Arc::StateId start = (&clat2)->Start();
            kaldi::BaseFloat total_backward_cost = beta[start];
            total_backward_cost -= Log(1.0f - lat_weight);
            for(fst::StateIterator<CompactLattice> sit(clat2); !sit.Done(); sit.Next()){
              CompactLatticeWeight f = (&clat2)->Final(sit.Value());
              LatticeWeight w = f.Weight();
              w.SetValue1(w.Value1() + total_backward_cost);
              f.SetWeight(w);
              (&clat2)->SetFinal(sit.Value(), f);
            }
          }
          else{
            fail = true; // Failed to compute backward probabilities on clat2.
          }
        }
      }
      else{
        fail = true; // Lattice weights are positive numbers, and Lattice 1 weight + Lattice 2 weight must be 1.0f.
      }
      if(fail){
        std::ostringstream ss;
        ss << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
        result_client_->broadcast(ss.str());
        return;
      }
      fst::Union(&clat1, clat2);
    }
    //fst::ConvertLattice(clat1, &rescored_lat);
    kaldi::MinimumBayesRisk mbr(clat1);
    words = mbr.GetOneBest();
    weight.SetValue1(-0.0f);
    weight.SetValue2(-0.0f);
  }
  else if(false){ // This mode just outputs the offline results of the second model.
    kaldi::Lattice best_path_lat;
    cmb_decoder_.GetBestPath(&best_path_lat);
    GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
    GetWordAlignmentSequence(best_path_lat, cmb_trans_model_, cmb_lexicon_info_, &words, &weight);
  }
  else{
    return;
  }

/*
  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned, &weight);
*/

  if(emulate_julius_){
    if(features_){
      ss << std::fixed << std::setprecision(6)
         <<   "source_id = " << features_->GetSourceInfo().source_id
         <<   ", azimuth = " << features_->GetSourceInfo().azimuth
         << ", elevation = " << features_->GetSourceInfo().elevation
         <<       ", sec = " << features_->GetSourceInfo().tv_sec
         <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
    }
    ss << "### Recognition: Rescoring pass \n"
       << "STAT: 00 \n"
       << "sentence1: ";
    if(word_syms_ != NULL){
      size_t i = 0;
      while(i < words.size()){
        std::string classid = word_syms_->Find(words[i]);
        ss << string_tokenizer(classid, "+")[0] << " ";
        if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
        i++;
      }
    }
    else{
      size_t i = 0;
      while(i < words.size()){ ss << words[i++] << " "; }
    }
    ss << "\n";
  }

  ss << (emulate_julius_ ? "wseq1: " : "result: ");
  if(word_syms_ != NULL){
    size_t i = 0;
    while(i < words.size()){
      std::string s = word_syms_->Find(words[i]);
      if(s == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
  }
  else{
    size_t i = 0;
    while(i < words.size()){ ss << words[i++] << " "; }
  }
  ss << "\n";

  ss << (emulate_julius_ ? "phseq1: " : "phone_seq: ");
  if(phone_syms_ != NULL){
    size_t i = 0;
    while(i < phones_aligned.size()){
      kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
      kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
      std::string s = phone_syms_->Find(phone);
      if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
      ss << s << " ";
      i++;
    }
  }
  ss << "\n";

  if(emulate_julius_){
    ss << "cmscore1: ";
    size_t i = 0;
    while(i < words.size()){
      ss << std::fixed << std::setprecision(3) << 1.0f << " ";
      i++;
    }
    ss << "\n";
    ss << std::fixed << std::setprecision(6)
       << "score1: " << -(double)(weight.Value1() + weight.Value2())
       << " ( AM: " << -(double)weight.Value2() << ", LM: " << -(double)weight.Value1() << " )";
  }
  else{
    ss << std::fixed << std::setprecision(6)
       <<     "score=" << -(double)(weight.Value1() + weight.Value2())
       << ", amscore=" << -(double)weight.Value2()
       << ", lmscore=" << -(double)weight.Value1();
  }
  out << ss.str() << std::endl;

  /*
  Lattice lat;
  decoder.GetRawLattice(&lat);
  CompactLattice clat;
  if(!DeterminizeLatticePhonePrunedWrapper(trans_model, &lat, decoder_config.lattice_beam, &clat, decoder_config.det_opts)){ }
  CompactLattice best_path_clat;
  CompactLatticeShortestPath(clat, &best_path_clat);

  Lattice best_path_lat;
  ConvertLattice(best_path_clat, &best_path_lat);
  */
}

} // namespace nnet3
} // namespace kaldi

